//
//  SimpleXMLRPC.m
//  SimpleXMLRPC
//
//  Created by Markus Hutzler on 10/8/11.
//  Copyright 2011 spiderware gmbh. All rights reserved.
//

#import "SimpleXMLRPC.h"
#import "NSData+Base64.h"

@implementation SimpleXMLRPC
@synthesize content = _content;
@synthesize url = _url;
@synthesize function = _function;
@synthesize timeout;


- (id)initWithString:(NSString *)content delegate:(id)delegate
{
    self = [super init];
    if (self) {
        _content = content;
        _delegate = delegate;
        _url = nil;
        [self parseContent];
    }
    return self;
}

- (id)initWithURL:(NSURL *)url function:(NSString *)function data:(id)data delegate:(id)delegate
{
    self = [super init];
    if (self) {
        _url = url;
        _function = function;
        _data = data;
        _delegate = delegate;
        timeout = 20.0;
        
    }
    return self;
}

#pragma mark - Parse Response

-(void) parseContent
{
    @try {
        scanner = [NSScanner scannerWithString:_content];
        if(![scanner scanString:@"<?xml version=\"1.0\"?>" intoString:nil])
            [self throwXMLRPCFormatException:@"header wrong" ];
        [self scanOrThrowException:@"<methodResponse>"];
        if([scanner scanString:@"<params>" intoString:nil])
            [_delegate simpleXMLRPC:self didFinishWithData:[self scanSuccess]];
            
        else if([scanner scanString:@"<fault>" intoString:nil])
            [_delegate simpleXMLRPC:self didFailWithData:[self scanFault]];
    }
    @catch (NSException *exception) {
        if ([[exception name] isEqualToString:@"XMLRPC not well formated!"] ) {
            //if ( [_delegate conformsToProtocol:@protocol(SimpleXMLRPCDelegate)] )
                [_delegate simpleXMLRPC:self didFailWithError:[[NSError alloc] initWithDomain:@"XMLRPC"
                                                                                     code:0 
                                                                                 userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[exception name],@"name",[exception reason],@"reason", nil]]];
        }
        else
            @throw;
    }

}


-(void) throwXMLRPCFormatException:(NSString *)reason
{
    [[NSException exceptionWithName:@"XMLRPC not well formated!" 
                             reason:[NSString stringWithFormat:@"%@ (at position %d)\n %@!ERROR!",reason,[scanner scanLocation],[[scanner string] stringByPaddingToLength:[scanner scanLocation] withString:[scanner string] startingAtIndex:0]] userInfo:nil] 
     raise];
    
}

- (void) scanOrThrowException:(NSString *)tag
{
    if(![scanner scanString:tag intoString:nil])
        [self throwXMLRPCFormatException:[NSString stringWithFormat:@"%@ tag expected",tag]];
}

-(id) scanSuccess
{
    [self scanOrThrowException:@"<param>"];
    
    id ret = [self scanValue];
    
    [self scanOrThrowException:@"</param>"];
    [self scanOrThrowException:@"</params>"];
    [self scanOrThrowException:@"</methodResponse>"];
    
    return ret;
}

-(id) scanFault
{
    id ret = [self scanValue];
    
    [self scanOrThrowException:@"</fault>"];
    [self scanOrThrowException:@"</methodResponse>"];
    
    return ret;    
}

-(NSString *) scanName
{
    if([scanner scanString:@"<name>" intoString:nil])
    {
        NSString * ret = [[[NSString alloc] init] autorelease];
        [scanner scanUpToString:@"<" intoString:&ret];
        [self scanOrThrowException:@"</name>"];
        
        return ret;
    }
    [self throwXMLRPCFormatException:@"<name> tag missing"];
    return nil; // will be never reached... avoid compiler warning
}


-(id) scanValue
{
    id ret = nil;
    [self scanOrThrowException:@"<value>"];
    if([scanner scanString:@"<int>" intoString:nil])
    {
        int v;
        [scanner scanInt:&v];
        ret =[NSNumber numberWithInt:v];
        [self scanOrThrowException:@"</int>"];
    }
    else if([scanner scanString:@"<i4>" intoString:nil])
    {
        int v;
        [scanner scanInt:&v];
        ret =[NSNumber numberWithInt:v];
        [self scanOrThrowException:@"</i4>"];
    }
    else if([scanner scanString:@"<boolean>" intoString:nil])
    {
        int v;
        [scanner scanInt:&v];
        ret =[NSNumber numberWithBool:v];
        [self scanOrThrowException:@"</boolean>"];
    }
    else if([scanner scanString:@"<double>" intoString:nil])
    {
        double v;
        [scanner scanDouble:&v];
        ret =[NSNumber numberWithDouble:v];
        [self scanOrThrowException:@"</double>"];
    }
    else if([scanner scanString:@"<string>" intoString:nil])
    {
        ret = [[NSString alloc] init];
        [scanner scanUpToString:@"<" intoString:&ret];
        [self scanOrThrowException:@"</string>"];
    }
    else if([scanner scanString:@"<dateTime.iso8601>" intoString:nil])
    {
        NSString * temp = [[NSString alloc] init];
        [scanner scanUpToString:@"<" intoString:&temp];
        NSDateFormatter* dateFormatter = [[NSDateFormatter alloc] init];
        [dateFormatter setDateFormat:@"yyyyMMdd'T'HH:mm:ss"];
        ret = [dateFormatter dateFromString:temp];
        [dateFormatter release];
        //[temp release];
        if (!ret) {
            [self throwXMLRPCFormatException:@"wrong date format in xml" ];
        }
        
        [self scanOrThrowException:@"</dateTime.iso8601>"];
    }
    else if([scanner scanString:@"<base64>" intoString:nil])
    {
        NSString * temp = [[NSString alloc] init];
        [scanner scanUpToString:@"<" intoString:&temp];
        ret = [NSData dataWithBase64:temp];
        
        [self scanOrThrowException:@"</base64>"];
    }
    else if([scanner scanString:@"<struct>" intoString:nil])
    {
        ret = [self scanDictionary];
    }
    else if([scanner scanString:@"<array>" intoString:nil])
    {
        ret = [self scanArray];
    }
    if (!ret) 
        [self throwXMLRPCFormatException:@"no valid value type found" ];
    
    [self scanOrThrowException:@"</value>"];    
    //NSLog(@"ret retaincount: %d %@",[ret retainCount],[[ret class] description]);
    return ret;
}

-(NSArray *)scanArray
{
    [self scanOrThrowException:@"<data>"];   
    NSMutableArray * ret = [[NSMutableArray alloc] init];
    while (![scanner scanString:@"</data>" intoString:nil])
        [ret addObject:[self scanValue]];
    
    [self scanOrThrowException:@"</array>"];
    
    return [(NSArray *)ret autorelease];
    
}

-(NSDictionary *)scanDictionary
{
    NSMutableDictionary * ret = [[NSMutableDictionary alloc] init];
    
    NSString * name;
    id v;
    while (![scanner scanString:@"</struct>" intoString:nil]) {
        [self scanOrThrowException:@"<member>"];
        name = [self scanName];
        v = [self scanValue];
        [self scanOrThrowException:@"</member>"];
        [ret setObject:v forKey:name];
        
    }
    return [(NSDictionary *)ret autorelease];
}

#pragma mark - Request Generation

-(NSString *) makeXMLOfObject:(id) o
{    
    if([o isKindOfClass:[NSString class]])
    {
        return [NSString stringWithFormat:@"<value><string>%@</string></value>",o];
    }
    else if([o isKindOfClass:[NSData class]])
    {
        return [NSString stringWithFormat:@"<value><base64>%@</base64></value>",[o base64]];
    }
    else if([o isKindOfClass:[NSDate class]])
    {
        NSDateFormatter* dateFormatter = [[[NSDateFormatter alloc] init] autorelease];
        [dateFormatter setDateFormat:@"yyyyMMdd'T'HH:mm:ss"];
        return [NSString stringWithFormat:@"<value><dateTime.iso8601>%@</dateTime.iso8601></value>",[dateFormatter stringFromDate:o]];
    }
    else if([o isKindOfClass:[NSArray class]])
    {
        NSMutableString * ret =[NSMutableString string];
        [ret appendString: @"<value><array><data>"];
        for (id x in o) {
            [ret appendString:[self makeXMLOfObject:x]];
        }
        [ret appendString:@"</data></array></value>"]; 
        return (NSString *)ret;
    }
    else if([o isKindOfClass:[NSDictionary class]])
    {
        NSMutableString * ret =[NSMutableString string];
        [ret appendString: @"<value><struct>"];
        for (NSString * key in [o allKeys]) {
            [ret appendString:@"<member>"];    
            [ret appendFormat:@"<name>%@</name>",key];    
            [ret appendString:[self makeXMLOfObject:[o objectForKey:key]]];
            [ret appendString:@"</member>"];    
        }
        [ret appendString:@"</struct></value>"];  
        return (NSString *)ret;
    }
    else if([[[o class] description] isEqualToString:@"NSCFBoolean"])
        return [NSString stringWithFormat:@"<value><bool>%d</bool></value>",[o boolValue]];
    else if([o isKindOfClass:[NSNumber class]])
    {
        if (!strncmp([o objCType], "i", 1)) {
            return [NSString stringWithFormat:@"<value><int>%d</int></value>",[o intValue]];
        }
        else if (!strncmp([o objCType], "d", 1) | !strncmp([o objCType], "f", 1)) 
            return [NSString stringWithFormat:@"<value><double>%f</double></value>",[o doubleValue]];
        [[NSException exceptionWithName:@"XMLRPC unknown Object!"  
                                 reason:[NSString stringWithFormat:@"can't convert object 'NSNumber' with objCType '%s' to XML.",[o objCType]] 
                               userInfo:nil] raise];
    }
    [[NSException exceptionWithName:@"XMLRPC unknown object!"  
                                reason:[NSString stringWithFormat:@"can't convert object '%@' to XML.",[[o class] description]] 
                               userInfo:nil] raise];
    return nil; // will be never reached... avoid compiler warning
}

-(NSString * ) makeHeader
{    
    return [NSString stringWithFormat: @"<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
                      "<methodCall>"
                      "<methodName>%@</methodName>"
                      "<params><param>%@</param></params>"
                      "</methodCall>",_function,[self makeXMLOfObject:_data]];    
}

-(void) call
{
    if (!_url)
        return; // no URL, no call!!!
    NSMutableURLRequest *request=[NSMutableURLRequest requestWithURL:_url
                                                            cachePolicy:NSURLRequestReloadIgnoringLocalAndRemoteCacheData
                                                        timeoutInterval:5.0];
    
    [request setValue:@"text/xml" forHTTPHeaderField: @"Content-Type"];
    [request setHTTPMethod: @"POST" ];
    [request setHTTPBody:[[self makeHeader] dataUsingEncoding:NSASCIIStringEncoding allowLossyConversion:YES]];
    
    //DEBUG:
    NSLog(@"%@",[[[NSString alloc] initWithData:[request HTTPBody] encoding:NSStringEncodingConversionAllowLossy] autorelease ]);
    NSURLConnection * connection = [[NSURLConnection alloc] initWithRequest:request delegate:self];
    if (connection) {
        receivedData = [[NSMutableData data] retain];
    } else {
        // some error???
    }
}

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response
{
    [receivedData setLength:0];
}
- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)d
{
    if ([receivedData length] == 0) {
        // first received package
        NSString * firstPackage = [[NSString alloc] initWithData:d encoding:NSStringEncodingConversionAllowLossy];
        if (![firstPackage hasPrefix:@"<?xml version=\"1.0\"?>"]) {
            // no valid XMLRPC Header!
            [receivedData release];
            [connection cancel];
            [connection release];
            [_delegate simpleXMLRPC:self didFailWithError:[NSError errorWithDomain:@"XMLRPC" code:0 userInfo:[NSDictionary dictionaryWithObjectsAndKeys:@"XMLRPC not well formated!",@"name",@"XML header missing.",@"reason", nil]]];
            return;
        }
        [firstPackage release];
        
    }
    [receivedData appendData:d];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)e
{
    [connection release];
    [receivedData release];
    [_delegate simpleXMLRPC:self didFailWithError:e];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection
{
    _content = [[NSString alloc] initWithData:receivedData encoding:NSStringEncodingConversionAllowLossy];
    [self parseContent];
    [_content release];
    [receivedData release];
    [connection release];
}

@end
